/*
 * This code is based on pucrunch 1.11 30-Jan-2000
 * by Pasi 'Albert' Ojala <albert@cs.tut.fi>.
 * See <URL:http://www.cs.tut.fi/%7Ealbert/Dev/pucrunch/>.
 */

#include<stdlib.h>
#include<string.h>
#include<assert.h>

#include"crunch.h"

static int maxGamma, escBits, escMask, extraLZPosBits, rleUsed;

#ifndef min
# define min(a,b) ((a<b)?(a):(b))
#endif

#define LRANGE		(((2<<maxGamma)-3)*256)	/* 0..125, 126 -> 1..127 */
#define MAXLZLEN	(2<<maxGamma)
#define MAXRLELEN	(((2<<maxGamma)-2)*256)	/* 0..126 -> 1..127 */

static int lrange, maxlzlen, maxrlelen;


/** work area for output buffer and header */
static uint8_t workArea[65536];
#define OUT_SIZE ((sizeof workArea) - (9 + 32))
static uint8_t* outBuffer;
static uint_fast16_t outPointer;
static uint_fast8_t bitMask;

static void
FlushBits (void)
{
  if (bitMask != 0x80)
    outPointer++;
}

/** Output a bit */
static void
PutBit (int bit)
{
  if (bit && outPointer < OUT_SIZE)
    outBuffer[outPointer] |= bitMask;
  bitMask >>= 1;
  if (!bitMask) {
    bitMask = 0x80;
    outPointer++;
  }
}

/** Output a number of the least significant bits of a value */
static void
PutNBits (int byte, int bits)
{
  while (bits--)
    PutBit ((byte & (1 << bits)));
}

/** Output a value */
static void
PutValue (int value)
{
  int bits = 0, count = 0;

  while (value > 1) {
    bits = (bits << 1) | (value & 1);	/* is reversed compared to value */
    value >>= 1;
    count++;
    PutBit (1);
  }
  if (count < maxGamma)
    PutBit (0);
  while (count--) {
    PutBit ((bits & 1));	/* output is reversed again -> same as value */
    bits >>= 1;
  }
}

static int
RealLenValue (int value)
{
  int count = 0;
  if (value < 2)	/* 1 */
    count = 0;
  else if (value < 4)	/* 2-3 */
    count = 1;
  else if (value < 8)	/* 4-7 */
    count = 2;
  else if (value < 16)	/* 8-15 */
    count = 3;
  else if (value < 32)	/* 16-31 */
    count = 4;
  else if (value < 64)	/* 32-63 */
    count = 5;
  else if (value < 128)	/* 64-127 */
    count = 6;
  else if (value < 256)	/* 128-255 */
    count = 7;

  if (count < maxGamma)
    return 2 * count + 1;
  return 2 * count;
}

static int lenValue[256];

static void
InitValueLen (void)
{
  int i;
  for (i = 1; i < 256; i++)
    lenValue[i] = RealLenValue (i);
}
#define LenValue(a) (lenValue[a])

/** Output a normal sequence */
static int
OutputNormal (int *esc, const uint8_t *data, int newesc)
{
  if ((data[0] & escMask) == *esc) {
    PutNBits ((*esc >> (8 - escBits)), escBits);	/* escBits>=0 */
    PutBit (0);
    PutBit (1);
    PutBit (0);
    *esc = newesc;
    PutNBits ((*esc >> (8 - escBits)), escBits);	/* escBits>=0 */
    PutNBits (data[0], 8 - escBits);
    return 1;
  }
  PutNBits (data[0], 8);
  return 0;
}

/** Output an end-of-file marker */
static void
OutputEof (int *esc)
{
  /* EOF marker */
  PutNBits ((*esc >> (8 - escBits)), escBits);	/* escBits>=0 */
  PutValue (3);	/* >2 */
  PutValue ((2 << maxGamma) - 1);	/* Maximum value */
  /* flush */
  FlushBits ();
}

static int rleHist[256];
static uint8_t rleValues[32];

static void
PutRleByte(int data)
{
  int index;
  for (index = 1; index < 32; index++) {
    if (data == rleValues[index]) {
      PutValue(index);
      return;
    }
  }
  PutValue (32 + (data >> 3));
  PutNBits (data, 3);
  /* Note: values 64..127 are not used if maxGamma>5 */
}

static unsigned char rleLen[256];
static void
InitRleLen (void)
{
  int i;
  for (i = 0; i < 256; i++)
    rleLen[i] = LenValue (32 + 0) + 3;
  for (i = 1; i < 32; i++)
    rleLen[rleValues[i]] = LenValue (i);
}
#define LenRleByte(d) (rleLen[d])

static int
LenRle (int len, int data)
{
  int out = 0;
  do {
    if (len == 1) {
      out += escBits + 3 + 8;
      len = 0;
    } else if (len <= (1 << maxGamma)) {
      out += escBits + 3 + LenValue (len-1) + LenRleByte (data);
      len = 0;
    } else {
      int tmp = min (len, maxrlelen);
      out += escBits + 3 + maxGamma + 8 +
	LenValue (((tmp - 1) >> 8) + 1) + LenRleByte (data);
      len -= tmp;
    }
  } while (len);
  return out;
}

static int
OutputRle (int *esc, const uint8_t *data, int rlelen)
{
  int len = rlelen;

  while (len) {
    if (len >= 2 && len <= (1<<maxGamma)) {
      PutNBits ((*esc >> (8 - escBits)), escBits);	/* escBits>=0 */
      PutBit (0);
      PutBit (1);
      PutBit (1);
      PutValue (len - 1);
      PutRleByte (*data);
      return 0;
    }
    if (len < 3) {
      while (len--)
	OutputNormal (esc, data, *esc);
      return 0;
    }

    if (len <= maxrlelen) {
      /* Run-length encoding */
      PutNBits ((*esc>>(8-escBits)), escBits);	/* escBits>=0 */

      PutBit (0);
      PutBit (1);
      PutBit (1);

      PutValue ((1 << maxGamma) + (((len - 1) & 0xff) >> (8 - maxGamma)));

      PutNBits ((len - 1), 8 - maxGamma);
      PutValue (((len - 1) >> 8) + 1);
      PutRleByte (*data);

      return 0;
    }

    /* Run-length encoding */
    PutNBits((*esc >> (8 - escBits)), escBits);	/* escBits>=0 */

    PutBit (0);
    PutBit (1);
    PutBit (1);

    PutValue ((1 << maxGamma) + (((maxrlelen - 1) & 0xff) >> (8 - maxGamma)));

    PutNBits ((maxrlelen - 1) & 0xff, 8 - maxGamma);
    PutValue (((maxrlelen - 1) >> 8) + 1);
    PutRleByte (*data);

    len -= maxrlelen;
    data += maxrlelen;
  }

  return 0;
}

static int
LenLz (int lzlen, int lzpos)
{
  if (lzlen == 2) {
    if (lzpos <= 256)
      return escBits + 10;
    else
      return 100000;
  }

  return escBits + 8 + extraLZPosBits +
    LenValue (((lzpos - 1) >> (8 + extraLZPosBits)) + 1) +
    LenValue (lzlen - 1);
}

static void
OutputLz (int *esc, int lzlen, int lzpos)
{
  assert (lzlen >= 2 && lzlen <= maxlzlen);
  PutNBits ((*esc >> (8 - escBits)), escBits);	/* escBits>=0 */
  PutValue (lzlen - 1);
  if (lzlen == 2) {
    PutBit (0);
    assert (lzpos <= 256);
  }
  else {
    PutValue (((lzpos - 1) >> (8 + extraLZPosBits)) + 1);
    PutNBits (((lzpos - 1) >> 8), extraLZPosBits);
  }
  PutNBits (((lzpos - 1) & 0xff) ^ 0xff, 8);
}

static unsigned short *rle, *elr, *lzlen, *lzpos;
static int *length;
static int inlen;
static const uint8_t *indata;
static uint8_t *mode, *newesc;
static unsigned short *backSkip;

enum MODE {
  LITERAL = 0,
  LZ77 = 1,
  RLE = 2,
  MMARK = 3
};

static int lzopt = 0;

/** Optimize the length of a match
 * @param optimize	flag: 1=optimize
 */
static int
OptimizeLength (int optimize)
{
  int i;

  length[inlen] = 0;		/* one off the end, our 'target' */
  for (i = inlen - 1; i >= 0; i--) {
    int r1 = 8 + length[i + 1], r2, r3;

    if (!lzlen[i] && !rle[i]) {
      length[i] = r1;
      mode[i] = LITERAL;
      continue;
    }

    /* If rle > MAXLZLEN, skip to the start of the rle - maxlzlen.. */
    if (rle[i] > maxlzlen && elr[i] > 1) {
      int z = elr[i];

      i -= elr[i];

      r2 = LenRle (rle[i], indata[i]) + length[i + rle[i]];
      if (optimize) {
	int ii, mini = rle[i], minv = r2;

	int bot = rle[i] - (1 << maxGamma);
	if (bot < 2)
	  bot = 2;

	for (ii = mini - 1; ii >= bot; ii--) {
	  int v = LenRle (ii, indata[i]) + length[i + ii];
	  if (v < minv) {
	    minv = v;
	    mini = ii;
	  }
	}
	if (minv != r2) {
	  lzopt += r2 - minv;
	  rle[i] = mini;
	  r2 = minv;
	}
      }
      length[i] = r2;
      mode[i] = RLE;

      for (; z >= 0; z--) {
	length[i + z] = r2;
	mode[i + z] = RLE;
      }
      continue;
    }
    r3 = r2 = r1 + 1000; /* r3 >= r2 > r1 */

    if (rle[i]) {
      r2 = LenRle(rle[i], indata[i]) + length[i + rle[i]];

      if (optimize) {
	int ii, mini = rle[i], minv = r2;

	/* Check only the original length and all shorter
	   lengths that are power of two.

	   Does not really miss many 'minimums' this way,
	   at least not globally..

	   Makes the assumption that the Elias Gamma Code is
	   used, i.e. values of the form 2^n are 'optimal' */
	ii = 2;
	while (rle[i] > ii) {
	  int v = LenRle (ii, indata[i]) + length[i + ii];
	  if (v < minv) {
	    minv = v;
	    mini = ii;
	  }
	  ii <<= 1;
	}

	if (minv != r2) {
	  lzopt += r2 - minv;
	  rle[i] = mini;
	  r2 = minv;
	}
      }
    }
    if (lzlen[i]) {
      r3 = LenLz (lzlen[i], lzpos[i]) + length[i + lzlen[i]];
      if (optimize && lzlen[i]>2) {
	int ii, mini = lzlen[i], minv = r3;
	int topLen = LenLz (lzlen[i], lzpos[i]) - LenValue (lzlen[i] - 1);

	/* Check only the original length and all shorter
	   lengths that are power of two.

	   Does not really miss many 'minimums' this way,
	   at least not globally..

	   Makes the assumption that the Elias Gamma Code is
	   used, i.e. values of the form 2^n are 'optimal' */
	ii = 4;
	while (lzlen[i] > ii) {
	  int v = topLen + LenValue(ii-1) + length[i + ii];
	  if (v < minv) {
	    minv = v;
	    mini = ii;
	  }
	  ii <<= 1;
	}

	/*
	  Note:
	  2-byte optimization checks are no longer done
	  with the rest, because the equation gives too long
	  code lengths for 2-byte matches if extraLzPosBits > 0.
	*/
	/* Two-byte rescan/check */
	if (backSkip[i] && backSkip[i] <= 256) {
	  /* There are previous occurrences (near enough) */
	  int v = LenLz (2, backSkip[i]) + length[i + 2];

	  if (v < minv) {
	    minv = v;
	    mini = 2;
	    lzlen[i] = mini;
	    r3 = minv;
	    lzpos[i] = (int) backSkip[i];
	  }
	}

	if (minv != r3 && minv < r2) {
	  lzopt += r3 - minv;
	  lzlen[i] = mini;
	  r3 = minv;
	}
      }
    }

    if (r2 <= r1) {
      if (r2 <= r3) {
	length[i] = r2;
	mode[i] = RLE;
      } else {
	length[i] = r3;
	mode[i] = LZ77;
      }
    } else {
      if (r3 <= r1) {
	length[i] = r3;
	mode[i] = LZ77;
      } else {
	length[i] = r1;
	mode[i] = LITERAL;
      }
    }
  }
  return length[0];
}

/*
    The algorithm in the OptimizeEscape() works as follows:
    1) Only unpacked bytes are processed, they are marked
       with MMARK. We proceed from the end to the beginning.
       Variable A (old/new length) is updated.
    2) At each unpacked byte, one and only one possible
       escape matches. A new escape code must be selected
       for this case. The optimal selection is the one which
       provides the shortest number of escapes to the end
       of the file,
	i.e. A[esc] = 1+min(A[0], A[1], .. A[states-1]).
       For other states A[esc] = A[esc];
       If we change escape in this byte, the new escape is
       the one with the smallest value in A.
    3) The starting escape is selected from the possibilities
       and mode 0 is restored to all mode 3 locations.

 */

static int
OptimizeEscape (int *startEscape, int *nonNormal)
{
  int i, j, states = 1 << escBits;
  int_fast32_t minp = 0, minv = 0, other = 0;
  int_fast32_t a[256];
  int_fast32_t b[256]; /* Remembers the # of escaped for each */
  int esc8 = 8 - escBits;

  if (states > 256)
    return 0;

  for (i = 256; i--; )
    b[i] = a[i] = -1;

  /* Mark those bytes that are actually output */
  for (i = 0; i < inlen; ) {
    switch (mode[i]) {
    case LZ77:
      other++;
      i += lzlen[i];
      break;
    case RLE:
      other++;
      i += rle[i];
      break;
    case LITERAL:
    default:
      mode[i++] = MMARK; /* mark it used so we can identify it */
      break;
    }
  }

  for (i = inlen - 1; i >= 0; i--) {
    /* Using a table to skip non-normal bytes does not help.. */
    if (mode[i] == MMARK) {
      int k = (indata[i] >> esc8);

      /* Change the tag values back to normal */
      mode[i] = LITERAL;

      /*
	k are the matching bytes,
	minv is the minimum value,
	minp is the minimum index
      */

      newesc[i] = minp << esc8;
      a[k] = minv + 1;
      b[k] = b[minp] + 1;
      if (k == minp) {
	/* Minimum changed -> need to find a new minimum */
	/* a[k] may still be the minimum */
	minv++;
	for (k = states - 1; k >= 0; k--) {
	  if (a[k] < minv) {
	    minv = a[k];
	    minp = k;
	    /*
	      There may be others, but the first one that
	      is smaller than the old minimum is equal to
	      any other new minimum.
	    */
	    break;
	  }
	}
      }
    }
  }

  /* Select the best value for the initial escape */
  if (startEscape) {
    i = inlen;	/* make it big enough */
    for (j = states - 1; j >= 0; j--) {
      if (a[j] <= i) {
	*startEscape = j << esc8;
	i = a[j];
      }
    }
  }
  if (nonNormal)
    *nonNormal = other;
  return b[startEscape ? (*startEscape >> esc8) : 0];
}

/* Initialize the RLE byte code table according to all RLE's found so far */
/* O(n) */
static void
InitRle (void)
{
  int p, mr, mv, i;

  for (i = 1; i < 32; i++) {
    mr = -1;
    mv = 0;

    for (p = 0; p < 256; p++) {
      if (rleHist[p] > mv) {
	mv = rleHist[p];
	mr = p;
      }
    }
    if (mv <= 0)
      break;
    rleValues[i] = mr;
    rleHist[mr] = -1;
  }
  InitRleLen ();
}


/* Initialize the RLE byte code table according to RLE's actually used */
/* O(n) */
static void
OptimizeRle (void)
{
  int p, mr, mv, i;

  memset (rleHist, 0, sizeof rleHist);

  for (p=0; p < inlen; ) {
    switch (mode[p]) {
    case LZ77: /* lz */
      p += lzlen[p];
      break;
    case RLE: /* rle */
      rleHist[indata[p]]++;
      p += rle[p];
      break;
    case LITERAL:
    case MMARK:
    default:
      p++;
      break;
    }
  }

  for (i = 1; i < 32; i++) {
    mr = -1;
    mv = 0;

    for (p = 0; p < 256; p++) {
      if (rleHist[p] > mv) {
	mv = rleHist[p];
	mr = p;
      }
    }
    if (mv <= 0)
      break;
    rleValues[i] = mr;
    rleHist[mr] = -1;
  }
  rleUsed = i - 1;
  InitRleLen ();
}

static void
OptimizeEscBits (int *startEscape)
{
  /*
    Absolute maximum number of escaped bytes with
    the escape optimize is 2^-n, where n is the
    number of escape bits used.

    This worst case happens only on equal-
    distributed normal bytes (01230123..).
    This is why the typical values are so much smaller.
  */

  int mb = 0, mv = 8 * OUT_SIZE;

  for (escBits = 1; escBits < 9; escBits++) {
    int escaped, other = 0, c;

    escMask = (0xff00 >> escBits) & 0xff;
    /* Find the optimum path for selected escape bits (no optimize) */
    OptimizeLength (0);
    /* Optimize the escape selections for this path & escBits */
    escaped = OptimizeEscape (startEscape, &other);

    /* Compare value: bits lost for escaping -- bits lost for prefix */
    if ((c = (escBits + 3) * escaped + other * escBits) >= mv)
      break; /* minimum found */
    mb = escBits;
    mv = c;
  }
  if (mb == 1) {	/* Minimum was 1, check 0 */
    int escaped;

    escBits = 0;
    escMask = 0;

    /* Find the optimum path for selected escape bits (no optimize) */
    OptimizeLength(0);
    /* Optimize the escape selections for this path & escBits */
    escaped = OptimizeEscape (startEscape, 0);
    if (3 * escaped < mv) {
      mb = 0;
      /* mv = 3*escaped; */
    }
  }

  escBits = mb;
  escMask = (0xff00 >> escBits) & 0xff;
}

static void
OptimizeLz (void)
{
  long lzstat[5] = { 0, 0, 0, 0, 0 }, i, cur = 0, old = extraLZPosBits;
  int p;

  for (p = 0; p < inlen; ) {
    switch (mode[p]) {
    case LZ77: /* lz */
      extraLZPosBits = 0;
      lzstat[0] += LenLz (lzlen[p], lzpos[p]);
      extraLZPosBits = 1;
      lzstat[1] += LenLz (lzlen[p], lzpos[p]);
      extraLZPosBits = 2;
      lzstat[2] += LenLz (lzlen[p], lzpos[p]);
      extraLZPosBits = 3;
      lzstat[3] += LenLz (lzlen[p], lzpos[p]);
      extraLZPosBits = 4;
      lzstat[4] += LenLz (lzlen[p], lzpos[p]);
      p += lzlen[p];
      break;
    case RLE: /* rle */
      p += rle[p];
      break;
    default: /* normal */
      p++;
      break;
    }
  }
  for (i = 5; --i; )
    if (lzstat[i] < lzstat[cur])
      cur = i;

  /* Find the optimum path (optimize) */
  if ((extraLZPosBits = cur) != old)
    OptimizeLength (1);
}

/** Compress a block.
 * @param gap	Length until memory "gap", or 0 if no gap.
 *		No back-references across the "gap" will be generated.
 * @return	the initial escape code
 */
static int
PackLz77 (int gap)
{
  int i, j, k, p;
  int escape, startEsc = -1;
  unsigned char *hashValue;
  const uint8_t *a;
  unsigned short *lastPair;

  InitRleLen ();
  length = calloc (sizeof *length, inlen + 1);
  mode = calloc (sizeof *mode, inlen);
  rle = calloc (sizeof *rle, inlen);
  elr = calloc (sizeof *elr, inlen);
  lzlen = calloc (sizeof *lzlen, inlen);
  lzpos = calloc (sizeof *lzpos, inlen);
  newesc = calloc (sizeof *newesc, inlen);
  backSkip  = calloc (sizeof *backSkip, inlen);
  hashValue = malloc (inlen);
  lastPair  = calloc (sizeof *lastPair, 256 * 256);

  /* error checking */
  if (!length || !mode || !rle || !elr || !lzlen || !lzpos || !newesc ||
      !lastPair || !backSkip || !hashValue)
    goto errorexit;

  i = 0;
  j = 0;
  a = indata + inlen;
  for (p = inlen - 1; p >= 0; p--) {
    k = j;
    j = i;
    i = *--a;	/* Only one read per position */

    /* Without hash: 18.56%, end+middle: 12.68% */
    /* hashValue[p] = i*2 ^ j*3 ^ k*5; */ /* 8.56% */
    /* hashValue[p] = i ^ j*2 ^ k*3; */   /* 8.85% */
    /* hashValue[p] = i + j + k; */       /* 9.33% */
    /* hashValue[p] = i + j*2 + k*3; */   /* 8.25% */
    /* hashValue[p] = i*2 + j*3 + k*5; */ /* 8.29% */
    /* hashValue[p] = i*3 + j*5 + k*7; */ /* 7.95% */
    hashValue[p] = i*3 + j*5 + k*7; /* 7.95 % */
  }

  /* Detect all RLE and LZ77 jump possibilities */
  for (p = 0; p < inlen; p++) {
    /* check run-length code - must be done, LZ77 search needs it! */
    if (rle[p] <= 0) {
      /*
	There are so few RLE's and especially so few
	long RLE's that byte-by-byte is good enough.
      */
      const uint8_t *a = indata + p;
      int val = *a++; /* if this were uchar, it would go to stack..*/
      int top = inlen - p;
      int rlelen = 1;

      /* Loop for the whole RLE */
      while (rlelen < top && *a++ == (unsigned char) val)
	rlelen++;

      if (rlelen >= 2) {
	rleHist[indata[p]]++;

	for (i=rlelen - 1; i >= 0; i--) {
	  rle[p + i] = rlelen - i;
	  elr[p + i] = i;	/* For RLE backward skipping */
	}
      }
    }

    /* check LZ77 code */
    if (p + rle[p] + 1 < inlen) {
      int bot = p - lrange, maxval, maxpos, rlep = rle[p];
      unsigned char hashCompare = hashValue[p];

      /*
	There's always 1 equal byte, although it may
	not be marked as RLE.
      */
      if (rlep <= 0)
	rlep = 1;
      if (bot < 0)
	bot = 0;
      bot += (rlep - 1);

      /*
	First get the shortest possible match (if any).
	If there is no 2-byte match, don't look further,
	because there can't be a longer match.
      */
      i = lastPair[(indata[p] << 8) | indata[p + 1]] - 1;
      if (i >= bot && (i >= gap || p + 1 < gap)) {
	/* Got a 2-byte match at least */
	maxval = 2;
	maxpos = p - i;

	/*
	  A..AB	rlep # of A's, B is something else..

	  Search for bytes that are in p + (rlep-1), i.e.
	  the last rle byte ('A') and the non-matching one
	  ('B'). When found, check if the rle in the compare
	  position (i) is long enough (i.e. the same number
	  of A's at p and i-rlep+1).

	  There are dramatically less matches for AB than for
	  AA, so we get a huge speedup with this approach.
	  We are still guaranteed to find the most recent
	  longest match there is.
	*/

	i = lastPair[(indata[p + (rlep - 1)] << 8) | indata[p + rlep]] - 1;
	while (i >= bot && (i >= gap || p + rlep < gap)) {
	  /* Equal number of A's ? */
	  if (!(rlep - 1) || rle[i - (rlep - 1)] == rlep) {
	    /* 'head' matches */
	    /*
	      Check the hash values corresponding to the last
	      two bytes of the currently longest match and
	      the first new matching(?) byte. If the hash
	      values don't match, don't bother to check the
	      data itself.
	    */
	    if (hashValue[i + maxval - rlep - 1] == hashCompare) {
	      const uint8_t *a = indata + i + 2;	/* match  */
	      const uint8_t *b = indata + p + rlep + 1;	/* curpos */
	      int topindex = inlen - (p + rlep - 1);

	      /* the 2 first bytes ARE the same.. */
	      j = 2;
	      while (j < topindex && *a++ == *b++)
		j++;

	      /* Avoid crossing the gap by limiting the match length. */
	      if (i < gap && p + rlep - 1 + j >= gap)
		j = gap - p - rlep;

	      if (j + rlep - 1 > maxval) {
		int tmplen = j + rlep - 1, tmppos = p - i + rlep - 1;

		if (tmplen > maxlzlen)
		  tmplen = maxlzlen;

		/* Accept only versions that really are shorter */
		if (tmplen * 8 - LenLz (tmplen, tmppos) >
		    maxval * 8 - LenLz (maxval, maxpos)) {
		  maxval = tmplen;
		  maxpos = tmppos;
		  hashCompare = hashValue[p + maxval - 2];
		}
		if (maxval == maxlzlen)
		  break;
	      }
	    }
	  }
	  if (!backSkip[i])
	    break; /* No previous occurrences (near enough) */
	  i -= backSkip[i];
	}

	/*
	  If there is 'A' in the previous position also,
	  RLE-like LZ77 is possible, although rarely
	  shorter than real RLE.
	*/
	if (p && rle[p - 1] > maxval) {
	  maxval = rle[p - 1] - 1;
	  maxpos = 1;
	}
	/*
	  Last, try to find as long as possible match
	  for the RLE part only.
	*/
	if (maxval < maxlzlen && rlep > maxval) {
	  bot = p - lrange;
	  if (bot < 0)
	    bot = 0;

	  /* Note: indata[p] == indata[p + 1] */
	  i = lastPair[indata[p] * 257] - 1;
	  while (i >= bot) {
	    if (elr[i] + 2 > maxval) {
	      maxval = min(elr[i] + 2, rlep);
	      maxpos = p - i + (maxval - 2);
	      if(maxval == rlep)
		break; /* Got enough */
	    }
	    i -= elr[i];
	    if (!backSkip[i])
	      break; /* No previous occurrences (near enough) */
	    i -= backSkip[i];
	  }
	}
	assert (maxval <= inlen - p);
	if (maxpos <= 256 || maxval > 2) {
	  assert (maxpos >= 0);
	  lzlen[p] = maxval < maxlzlen ? maxval : maxlzlen;
	  lzpos[p] = maxpos;
	}
      }
    }

    /* Update the two-byte history ('hash table') &
       backSkip ('linked list') */
    if (p + 1 < inlen) {
      int index = (indata[p] << 8) | indata[p + 1];
      int ptr = p - (lastPair[index] - 1);

      if (ptr > p || ptr > 0xffff)
	ptr = 0;

      backSkip[p] = ptr;
      lastPair[index] = p+1;
    }
  }

  /* Initialize the RLE selections */
  InitRle ();
  /* Check the normal bytes / all ratio */
  OptimizeEscBits (&escape);
  /* Find the optimum path (optimize) */
  OptimizeLength (1);
  OptimizeLz ();
  /* Optimize the escape selections */
  OptimizeEscape (&escape, 0);
  startEsc = escape;
  OptimizeRle ();	/* Retune the RLE selections */

  for (p = 0; p < inlen; ) {
    switch (mode[p]) {
    case LITERAL: /* normal */
      length[p] = outPointer;
      OutputNormal (&escape, indata + p, newesc[p]);
      p++;
      break;
    case LZ77: /* lz77 */
      /* Re-search matches to get the closest one */
      if (lzopt && /* If any changes to lengths.. */
	  lzlen[p] > 2 && lzlen[p] > rle[p]) {
	int bot = p - lzpos[p] + 1, i;
	unsigned short rlep = rle[p];

	if (!rlep)
	  rlep = 1;
	if (bot < 0)
	  bot = 0;
	bot += (rlep-1);

	i = p - backSkip[p];
	while (i >= bot) {
	  /* Equal number of A's ? */
	  if (rlep == 1 || rle[i - rlep + 1]==rlep) {
	    /* 'head' matches */
	    const uint8_t *a = indata + i + 1;		/* match  */
	    const uint8_t *b = indata + p + rlep;	/* curpos */
	    int topindex = inlen - (p + rlep - 1);

	    j = 1;
	    while (j < topindex && *a++ == *b++)
	      j++;

	    if (j + rlep - 1 >= lzlen[p]) {
	      int tmppos = p - i + rlep - 1;
	      lzpos[p] = tmppos;
	      break;
	    }
	  }
	  if (!backSkip[i])
	    break; /* No previous occurrences (near enough) */
	  i -= backSkip[i];
	}
      }

      for (i = 0; i < lzlen[p]; i++)
	length[p + i] = outPointer;
      OutputLz (&escape, lzlen[p], lzpos[p]);
      p += lzlen[p];
      break;

    case RLE: /* rle */
      for (i=0; i < rle[p]; i++)
	length[p + i] = outPointer;
      OutputRle (&escape, indata + p, rle[p]);
      p += rle[p];
      break;

    default: /* Error Flynn :-) */
      p++;
      assert (0);
      break;
    }
  }
  OutputEof (&escape);

 errorexit:
  if (rle)
    free (rle);
  if (elr)
    free (elr);
  if (lzlen)
    free (lzlen);
  if (lzpos)
    free (lzpos);
  if (length)
    free (length);
  if (mode)
    free (mode);
  if (newesc)
    free (newesc);
  if (lastPair)
    free (lastPair);
  if (backSkip)
    free (backSkip);
  if (hashValue)
    free (hashValue);
  return startEsc;
}

/** Initialize the static variables for the cruncher */
static void
InitPack ()
{
  maxGamma = 7;
  escBits = 2;
  escMask = (0xff00 >> escBits) & 0xff;
  extraLZPosBits = 0;
  rleUsed = 31;
  outPointer = 0;
  outBuffer = workArea + (9 + 32);
  bitMask = 0x80;
  memset (workArea, 0, sizeof workArea);
  memset (rleValues, 0, sizeof rleValues);
  rleValues[0] = 1;
  lrange = LRANGE;
  maxrlelen = MAXRLELEN;
  maxlzlen = MAXLZLEN;
  InitValueLen ();
}

/** Crunch a program
 * @param start		start address of the program
 * @param warp		warp page of the program
 * @param src		the input buffer
 * @param size		(input/output) size of the buffer
 * @return		the output buffer (NULL on failure)
 */
const uint8_t*
crunch (uint_fast16_t start,
	uint_fast8_t warp,
	const uint8_t* src,
	uint_fast16_t* size)
{
  uint8_t* dest;
  int escape;
  uint_fast16_t gap = (uint_fast16_t) warp << 8;

  if (gap > start)
    gap -= start;
  else
    gap = 0;

  InitPack ();
  indata = src;
  inlen = *size;
  if ((escape = PackLz77 (gap)) == -1)
    return 0;

  memcpy (dest = outBuffer - rleUsed, rleValues + 1, rleUsed);
  *--dest = rleUsed;
  *--dest = extraLZPosBits;
  *--dest = 1 << maxGamma;
  *--dest = maxGamma + 1;
  *--dest = escBits;
  *--dest = start >> 8;
  *--dest = start & 0xff;
  *--dest = escape >> (8 - escBits);
  *--dest = warp;
  *size = 9 + rleUsed + outPointer;
  return dest;
}
