;;; @file flash4.s
;;; Flash programming software for the Vic Flash Plugin
;;; @author Marko Mäkelä (marko.makela@iki.fi)

;;; This file can be compiled with xa
;;; (Cross-Assembler 65xx V2.1.4h 12dec1998 (c) 1989-98 by A.Fachat)
;;; or xa (xa65) v2.3.5
;;; Written by Andre Fachat, Jolse Maginnis, David Weinehall and Cameron Kaiser

;;; Copyright © 2003,2010 Marko Mäkelä (marko.makela@iki.fi)
;;;
;;;     This program is free software; you can redistribute it and/or modify
;;;     it under the terms of the GNU General Public License as published by
;;;     the Free Software Foundation; either version 2 of the License, or
;;;     (at your option) any later version.
;;;
;;;     This program is distributed in the hope that it will be useful,
;;;     but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;     GNU General Public License for more details.
;;;
;;;     You should have received a copy of the GNU General Public License
;;;     along with this program; if not, write to the Free Software
;;;     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

flash	= $a000			; base address of the flash ROM (BLK5)
bankrg	= $9800			; bank selection register
cfgrg	= $9801			; configuration register (default value $40)
	;; b7 == 1 => I/O2 disabled until RESET
	;; b6 == 1 => ROM write protect (default)
	;; b5 == 1 => RAM at BLK5 (instead of ROM)
	;; b4 => 0=3k (RAM1-3), 1=8k (BLK1)
	;; b3 == 1 => master RAM enable (BLK2,BLK3 is always enabled)
	;; b2,b1 == 0 on read ("inc cfgrg" assumes b2==0)
	;; b0 => A21

;;; the address lines A0 and A1 are swapped on the flash chip
f000	= flash			; flash address 0
f001	= flash + 2		; flash address 1
f555	= flash + $556		; flash address $555
f2aa	= flash + $2a9		; flash address $2aa

size	= $a8			; size of block being programmed (16 bits)
crsrchr	= $d1			; cursor position in screen memory
crsr_x	= $d3			; current cursor column
crsrcol	= $f3			; cursor position in color memory
txtcolor= 646			; current text color
pout	= $fb			; pointer to flash address (2 bytes)
pin	= $fd			; pointer to source address (2 bytes)
tmp	= $ff			; auxiliary memory
strout	= $cb1e			; output a NUL-terminated string
chrout	= $ffd2			; output a character

cTEXT	= 1			; white text color
cPROG	= 5			; green progress indicator color
cOK	= 0			; black screen signifies OK color
cBUSY	= 1			; white border signifies a wait
cERR	= 2			; red border signifies error
color	= $900f			; color register

;;; set the screen color to c
#define setcolor(c) lda #8+c:sta color
;;; print a NUL-terminated message
#define printmsg(x) lda#<x:ldy#>x:jsr strout
#define printmsg_rts(x) lda#<x:ldy#>x:jmp strout

	.word $1201
	*=$1201
prg	.word nextln
	.word 2010
	.byte $9e		; SYS
	.byte $30 + (start / 1000)
	.byte $30 + ((start - (start / 1000 * 1000)) / 100)
	.byte $30 + ((start - (start / 100 * 100)) / 10)
	.byte $30 + (start - (start / 10 * 10))
	.byte 0
nextln	.word 0
	.word ops		; pointer to the operation script
start	sei
	lda bankrg
	sta bankrg_save
	lda cfgrg
	sta cfgrg_save
	setcolor(cOK)
	lda #cTEXT
	sta txtcolor
	lda #$18
	sta cfgrg		; flash write enable, BLK1 enable
	lda #0
	sta bankrg
	ldy #$f0
	sty flash		; poke x, $f0 (read array data)
	ldx #$90		; $90 (autoselect)
	jsr command
	lda f000		; read 0 (manufacturer id)
	cmp #1
	beq mfr_ok

	sty flash		; poke x, $f0 (read array data)
	pha
	printmsg(mfr_mismatch)
	pla
	jsr hex2
	jmp error_println_exit

mfr_ok
	lda f001		; read 1 (device id)
	sty flash		; poke x, $f0 (read array data)
	cmp #$41
	beq dev_ok

	pha
	printmsg(dev_mismatch)
	pla
	jsr hex2
	jmp error_println_exit

dev_ok
	lda #<ops
	sta pin
	lda #>ops
	sta pin+1
	ldx #0
	lda (pin,x)		; read the first opcode without preincrement
	jsr op_do
oploop	ldx #0
	jsr incpinr
	jsr op_do
	jmp oploop
	;; read the next byte at the input position (.X must be 0)
incpinr	inc pin
	bne readpin
	inc pin+1
	bpl readpin
	printmsg(input_overflow)
	.byte 2			; hang to prevent further corruption
readpin	lda (pin,x)
	rts
	;; execute the operation in .A
op_do	pha
	lda pin+1
	jsr hex2
	lda pin
	jsr hex2
	lda #$20
	jsr chrout
	pla
	pha
	jsr hex2
	lda #$20
	jsr chrout
	pla
	cmp #(ops-optab)/2
	bcs op_unknown
	asl
	tax
	lda optab+1,x
	pha
	lda optab,x
	pha
	rts
op_unknown
	printmsg(unknown_op)
	pla			; discard the return address
	pla			; (return to caller's caller)
error_println_exit
	lda #$0d		; output a carriage return (line break)
	jsr chrout
	setcolor(cERR)		; flag error status
exit	lda #0			; exit
bankrg_save = *-1
	sta bankrg
	lda #$40
cfgrg_save = *-1
	sta cfgrg
	cli
	rts
	;; program data
program	ldx #0
	jsr incpinr		; parameter => size (16 bits)
	sta size
	jsr incpinr
	sta size+1
	jsr incpinr		; parameter => address (24 bits, 22 used)
	sta pout
	jsr incpinr
	sta pout+1
	jsr incpinr
	sta tmp
	lda pout+1
	asl
	rol tmp
	asl
	rol tmp
	asl
	rol tmp
	lda #$18		; flash write enable, BLK1 enable
	adc #0			; set address line A21 if needed
	sta cfgrg
	lda tmp
	sta bankrg
	lda pout+1
	and #$1f
	ora #>flash
	sta pout+1
	setcolor(cBUSY)
	printmsg(programming)
	lda size+1
	jsr hex2
	lda size
	jsr hex2
	printmsg(programming_size)
	jsr printaddr
	printmsg(programming_size_end)
	inc size
	inc size+1
	ldx #0
prog	clc
	dec size		; all bytes programmed?
	bne progn
	dec size+1
	beq op_complete
progn	jsr incpinr
	tay
	eor (pout,x)
	beq progok		; already programmed (skip)
	ldx #$a0
	jsr command		; command (write byte)
	tya
	ldx #0
	sta (pout,x)		; We cannot use post-indexed mode here,
				; because it would read from the address first.
ppoll	lda f000		; poll the operation status
	eor f000
	and #$44		; bits 2 and 6 ($44) toggling?
	eor #$44
	beq ppoll		; yes, keep polling
	lda #$f0		; switch to reading array data
	sta flash		; poke x, $f0 (read array data)
	tya
	cmp (pout,x)		; verify the data
	bne verror
progok	inc pout		; increment the address, wrap if necessary
	bne prog
	inc pout+1
	bit pout+1
	bvc prog
	lda #>flash		; wrap the bank at $c000
	sta pout+1
	inc bankrg
	bne prog
	inc cfgrg
	bvs prog		; branch always
op_all_complete
	pla			; discard the return address
	pla			; (return to caller's caller)
op_complete
	setcolor(cOK)
	printmsg_rts(operation_complete)

verror	pla			; discard the return address
	pla			; (return to caller's caller)
	printmsg(verify_failure)
	jsr printaddr
	printmsg(verify_failure_wrote)
	ldx #0
	lda (pin,x)
	jsr hex2
	printmsg(verify_failure_read)
	ldx #0
	lda (pout,x)
	jsr hex2
	jmp error_println_exit

erase	printmsg(erase_in_progress)
	ldx #$80		; command (erase)
	jsr command
	ldx #$10		; parameter (erase chip)
	jsr command
erasepoll
	setcolor(cBUSY)
	lda #cPROG
	ldy crsr_x
	sta (crsrcol),y		; set progress indicator color
epollx	ldx #progress_end - progress - 1
epoll	lda f000		; poll the device until operation complete
	eor f000
	and #$44		; bits 2 and 6 ($44) toggling?
	eor #$44
	bne erased		; no, completed
	lda progress,x
	sta (crsrchr),y
	dex
	bpl epoll
	bmi epollx
	ldy #$f0
	sty flash		; poke x, $f0 (read array data)
erased	ldx flash
	cpx #$ff
	beq op_complete
	pla			; discard the return address
	pla			; (return to caller's caller)
	txa
	pha
	printmsg(erase_failure)
	pla
	jsr hex2
	jmp error_println_exit

	;; erase a block
erase_blk
	printmsg(erase_sec_in_progress)
	jsr incpinr
	pha
	jsr hex2
	pla
	asl
	asl
	asl
	sta bankrg		; select the sector address
	lda #$18		; flash write enable, BLK1 enable
	adc #0			; set A21 if needed
	sta cfgrg
	printmsg(erase_sec_in_progress_at)
	lda #0
	sta pout
	sta pout+1
	jsr printaddr
	printmsg(erase_in_progress2)
	ldx #$80		; command (erase)
	jsr command
	ldx #$30		; parameter (sector, specified with A21..A16)
	jsr command
	bne erasepoll		; branch always

	;; print the current flash ROM address
printaddr
	lda pout+1
	asl
	asl
	asl
	sta tmp
	lda cfgrg		; read the most significant bit of the address
	lsr			; to the .C flag
	lda bankrg		; read the 8 next significant bits
	ror			; and shift 3 bits to the right
	ror tmp
	ror
	ror tmp
	ror
	ror tmp
	jsr hex2
	lda tmp
	jsr hex2
	lda pout
	;; print two hexadecimal digits in .A
hex2
	tax
	lsr
	lsr
	lsr
	lsr
	jsr hex1
	txa
	;; print one hexadecimal digit in .A
hex1
	and #$f
	ora #$30
	cmp #$3a
	bcc doprint
	adc #$41-$3a-1
doprint
	jmp chrout

	;; issue a command in .X, trash .A
command	lda #$aa
	sta f555
	lda #$55
	sta f2aa
	stx f555		; poke $555, command
	rts
	;; progress indicator
progress
	.byte 109,112,110,125
progress_end

mfr_mismatch
	.byte "EXPECTED MANUFACTURER", 13, "ID 01, GOT ", 0
dev_mismatch
	.byte "EXPECTED DEVICE", 13, "ID 41, GOT ", 0
unknown_op
	.byte 28, "UNKNOWN OP", 5, 0
input_overflow
	.byte "INPUT BUFFER OVERFLOW", 13, 0
erase_sec_in_progress
	.byte "ERASE SEC ", 0
erase_sec_in_progress_at
	.byte 13, "AT ", 0
erase_in_progress
	.byte "ERASE"
erase_in_progress2
	.byte ": ", 0
operation_complete
	.byte 30, "OK", 5, 13, 0
erase_failure
	.byte 28, "FAILED", 5, 13, "FIRST CELL IS ", 0
programming
	.byte "WRITE ", 0
programming_size
	.byte " B", 13, "TO ", 0
programming_size_end
	.byte ": ", 0
verify_failure
	.byte 28, "FAILED", 5, 13, "AT ", 0
verify_failure_wrote
	.byte 13, "WROTE: ", 0
verify_failure_read
	.byte " READ: ", 0

	;; jump table of operation codes
optab	.word op_all_complete-1, erase-1, erase_blk-1, program-1
	;; operations to be performed (terminated by 0)
ops
;	.byte 1			; erase chip

;	.byte 2			; erase sector
;	.byte 0			; parameter => sector number (0..63, A21..A16)

;	.byte 3			; write data
;	.word image_end-image_start
;	.byte 0, 0, 0		; at this address (little endian)
;#include "menu.s"		; the data (defines image_start and image_end)
	.byte 0			; end of operations
