#include<assert.h>
#include<search.h>
#include<stdio.h>
#include<string.h>
#include<stdlib.h>
#include<ctype.h>
#include"crunch.h"
#include"readline.h"

typedef uint_fast32_t addr_t;
typedef uint_fast16_t word_t;
typedef uint_fast8_t byte_t;
typedef uint8_t octet_t;

/** Flags for the "previous entry" pointer */
enum prevflags
{
  /** BASIC program */
  mBASIC = 0x40 << 16,
  /** Cartridge image */
  mCartridge = 0x80 << 16,
  /** Subdirectory */
  mDirectory = 0xc0 << 16
};
/** Flags for the "next entry" pointer: type of memory expansion */
enum nextflags
{
  /** 3k memory expansion */
  m3k = 0x40 << 16,
  /** 8k (up to 24k) memory expansion */
  m8k = 0x80 << 16,
  /** No memory expansion, or subdirectory */
  m0k = 0xc0 << 16
};

/** The name part of a menu entry */
typedef octet_t name_t[16];

/** Signifies an unknown memory address */
enum { ADDR_UNKNOWN = (1 << 22) - 1 };

/** Menu */
struct menu
{
  /** NULL-terminated singly linked list of unresolved references to
  the menu (NULL if the menu has been resolved) */
  struct menu* next;
  /** Line number of the reference or definition */
  unsigned lineno;
  /** If next==NULL, start offset, or ADDR_UNKNOWN if the menu is unresolved.
  If next != NULL, a menu referring to the entry. */
  addr_t addr;
};
/** Menu definition or reference */
struct menuref
{
  /** Menu label */
  char* name;
  /** The menu definition, or unresolved references to the menu */
  struct menu* data;
};

/** Menu entry */
struct menuent
{
  /** Previous entry, possibly ORed with enum prevflags */
  octet_t prev[3];
  /** Next entry, possibly ORed with enum nextflags */
  octet_t next[3];
  /** Entry text */
  name_t name;
};

/** first and last available byte on cartridge */
static addr_t firstAddr, lastAddr = ADDR_UNKNOWN + 1;
/** address of previous menu entry */
static addr_t menuPrev;

/** cartridge image file */
static FILE* cart;
/** offset to cartridge image */
static long cartoffset;

/** Search for the first available location
 * @return	0 on success, name of failed library call on failure
 */
static const char*
searchLast (void)
{
  /** right after the a0CBM signature */
  addr_t addr = 9;

  for (;;) {
    /** Location of next memory block (0xffffff on end of memory) */
    octet_t buf[3];
    addr_t next;
    if (fseek (cart, cartoffset + addr, SEEK_SET))
      return "fseek";
    if (1 != fread (&buf, sizeof buf, 1, cart))
      return "fread";
    next = (addr_t) buf[0] | (addr_t) buf[1] << 8 | (addr_t) buf[2] << 16;
    if (next == 0xffffff)
      break;
    if (addr >= next)
      return "corruption noticed";
    addr = next;
  }

  firstAddr = addr + 3;
  return 0;
}

/** Open a cartridge image
 * @param name	image file name
 * @return	the cartridge file pointer (or NULL if failed)
 */
static FILE*
openCartridge (const char* name)
{
  uint8_t buf[17];
  const size_t cart_len = 1 << 22;
  /** auto-start signature */
  static const unsigned char a0CBM[5] = { 0x41, 0x30, 0xc3, 0xc2, 0xcd };
  const char* s;

  if (!(cart = fopen (name, "r+b"))) {
    fputs (name, stderr);
    perror (": fopen");
    return 0;
  }

  if (sizeof buf != fread (buf, 1, sizeof buf, cart)) {
    fputs (name, stderr);
    perror (": fread");
    fclose (cart);
    return cart = 0;
  }

  if (!memcmp (buf + 6, a0CBM, 5))
    cartoffset = 2;
  else if (!memcmp (buf + 4, a0CBM, 5))
    cartoffset = 0;
  else {
  unrecognized:
    fputs (name, stderr);
    fputs (": not recognized as a cartridge image\n", stderr);
    fclose (cart);
    return cart = 0;
  }

  if (buf[cartoffset + 1] < 0xa0 ||
      buf[cartoffset + 1] > 0xbf ||
      buf[cartoffset + 3] < 0xa0)
    goto unrecognized;

  firstAddr =
    (uint32_t) buf[cartoffset + 12] |
    (uint32_t) buf[cartoffset + 13] << 8 |
    (uint32_t) buf[cartoffset + 14] << 16;
  if (firstAddr >= lastAddr)
    goto unrecognized;

  if (fseek (cart, 0, SEEK_END))
    s = "fseek";
  else {
    long l = ftell (cart);
    if (l == (long) cart_len) {
      /* The length is correct; do nothing */
    }
    else if (l <= 8192) {
      octet_t buf[BUFSIZ];
      size_t pad;
      /* Pad with 0xff to 4 megabytes */
      memset (buf, 0xff, sizeof buf);
      pad = BUFSIZ - (l % BUFSIZ);
      if (pad != fwrite (buf, 1, pad, cart)) {
      fwrite_failure:
	s = "fwrite";
	goto done;
      }
      l += pad;
      do {
	if (BUFSIZ != fwrite (buf, 1, BUFSIZ, cart))
	  goto fwrite_failure;
	l += BUFSIZ;
      }
      while (l < cart_len);
    }
    else {
      fputs (name, stderr);
      fputs (": should be 4194304 bytes or less than 8192 bytes\n", stderr);
      fclose (cart);
      cart = 0;
      return cart;
    }

    s = searchLast ();
  }
done:
  if (s) {
    fputs (name, stderr);
    fputs (": ", stderr);
    perror (s);
    fclose (cart);
    cart = 0;
  }

  return cart;
}

/** start and end address of the program */
static uint_fast16_t prgStart, prgEnd;
/** program buffer */
static uint8_t prgData[65536];
/** warp page of the program */
static uint_fast8_t prgWarp;
/** Flags for the "previous entry" pointer */
static enum prevflags prgPrevFlags;
/** Flags for the "next entry" pointer: type of memory expansion */
static enum nextflags prgNextFlags;

/** Encode an address
 * @param buf	buffer of 3 octets
 * @param addr	the address
 */
static void
encodeAddr (octet_t* buf,
	    addr_t addr)
{
  assert (addr < (1 << 24));

  buf[0] = addr;
  buf[1] = addr >> 8;
  buf[2] = addr >> 16;
}

/** Update an address in a menu entry
 * @param dest	where to write the address
 * @param addr	the address to be written
 * @return	0 on success, nonzero on failure
 */
static int
updateOffset (addr_t dest,
	      addr_t addr)
{
  octet_t buf[3];

  assert (dest + 3 <= lastAddr);

  if (fseek (cart, cartoffset + dest, SEEK_SET)) {
    perror ("fseek1(updateOffset)");
    return -1;
  }
  if (sizeof buf != fread (buf, 1, sizeof buf, cart)) {
    perror ("fread(updateOffset)");
    return -2;
  }
  if (fseek (cart, cartoffset + dest, SEEK_SET)) {
    perror ("fseek2(updateOffset)");
    return -3;
  }

  encodeAddr (buf, ((addr_t) (buf[2] & 0xc0) << 16) | addr);
  if (sizeof buf != fwrite (buf, 1, sizeof buf, cart)) {
    perror ("fwrite1(updateOffset)");
    return -4;
  }

  return 0;
}

/** Write a submenu reference
 * @param ent	the menu item
 * @param addr	the submenu offset (ADDR_UNKNOWN if unknown)
 * @return	offset of the reference, ADDR_UNKNOWN on failure
 */
static addr_t
writeSubmenu (struct menuent* ent,
	      addr_t addr)
{
  addr_t nextAddr = firstAddr + 3 + sizeof *ent;

  if (nextAddr > lastAddr)
    fputs ("no space left in image\n", stderr);
  else {
    encodeAddr (ent->prev, menuPrev | mDirectory);
    encodeAddr (ent->next, ADDR_UNKNOWN | m0k);

    /* Adjust the previous menu entry pointer */
    if (menuPrev != ADDR_UNKNOWN && updateOffset (menuPrev + 3, firstAddr));
    else if (fseek (cart, cartoffset + firstAddr, SEEK_SET))
      perror ("fseek3(writeSubmenu)");
    else if (sizeof *ent != fwrite (ent, 1, sizeof *ent, cart) ||
	     putc (addr & 0xff, cart) == EOF ||
	     putc ((addr >> 8) & 0xff, cart) == EOF ||
	     putc (addr >> 16, cart) == EOF)
      perror ("fwrite2(writeSubmenu)");
    else {
      menuPrev = firstAddr;
      firstAddr = nextAddr;
      return nextAddr - 3;
    }
  }
  return ADDR_UNKNOWN;
}

/** Close a menu entry (crunch the data)
 * @param ent	the menu item
 * @return	0 on success, nonzero on failure
 */
static int
closeEntry (struct menuent* ent)
{
  uint_fast16_t size = prgEnd - prgStart;
  const uint8_t* data = crunch (prgStart, prgWarp, prgData, &size);
  prgStart = prgEnd = 0;
  prgWarp = 0x80;
  if (!data)
    fputs ("crunch failed\n", stderr);
  else if (size > lastAddr - firstAddr - sizeof *ent)
    fputs ("no space left in image\n", stderr);
  else {
    addr_t nextAddr = firstAddr + size + sizeof *ent;

    encodeAddr (ent->prev, menuPrev | prgPrevFlags);
    encodeAddr (ent->next, ADDR_UNKNOWN | prgNextFlags);

    if (menuPrev != ADDR_UNKNOWN && updateOffset (menuPrev + 3, firstAddr));
    else if (fseek (cart, cartoffset + firstAddr, SEEK_SET))
      perror ("fseek(closeEntry)");
    else if (sizeof *ent != fwrite (ent, 1, sizeof *ent, cart) ||
	     size != fwrite (data, 1, size, cart))
      perror ("fwrite(closeEntry)");
    else {
      menuPrev = firstAddr;
      firstAddr = nextAddr;
      return 0;
    }
  }
  return -1;
}

/** Add a file to a menu entry
 * @param name		name of the file
 * @param lineno	input line number for diagnostics
 * @param flags		file flags
 * @param ent		the menu entry
 * @return		ent on success, NULL on failure
 */
static struct menuent*
addFile (const char* name,
	 unsigned lineno,
	 struct menuent* ent)
{
  /** length of the block read */
  uint_fast16_t length;
  /** room for start address */
  uint8_t start[2];
  /** the file */
  FILE* f = fopen (name, "rb");
  if (!f) {
    fprintf (stderr, "%u: %s: ", lineno, name);
    perror ("fopen");
    return 0;
  }
  if (2 != fread (start, 1, sizeof start, f)) {
    fprintf (stderr, "%u: %s: ", lineno, name);
    perror ("fread(start)");
    fclose (f);
    return 0;
  }
  length = prgEnd - prgStart;
  length = fread (prgData + length, 1, (sizeof prgData) - length, f);
  if (!length || ferror (f)) {
    fclose (f);
    fprintf (stderr, "%u: %s: ", lineno, name);
    perror ("fread(data)");
    return 0;
  }

  fclose (f);

  if (!prgStart && !prgEnd) {
    /* first file */
    prgStart = start[0];
    prgStart |= start[1] << 8;

    switch (prgStart) {
    case 0x0401: prgNextFlags = m3k; break;
    case 0x1001: prgNextFlags = m0k; break;
    case 0x1201: prgNextFlags = m8k; break;
    case 0x2000: case 0x3000:
      prgNextFlags = m8k;
      /* fall through */
    case 0x4000: case 0x5000: case 0x6000: case 0x7000:
    case 0xa000: case 0xb000:
      prgPrevFlags = mCartridge;
      break;
    default:
      fprintf (stderr, "%u: %s: unsupported start address 0x%04x\n",
	       lineno, name, prgStart);
      prgStart = 0;
      return 0;
    }

    prgEnd = length + prgStart;
    prgWarp = prgPrevFlags == mCartridge && prgEnd < 0x8000
      ? (prgEnd + 0xff) >> 8
      : 0x80;
    return ent;
  }
  else if ((!start[0] && start[1] == (uint8_t) 0xa0 &&
	    prgEnd == (uint_fast16_t) prgWarp << 8) ||
	   prgEnd == (((uint_fast16_t) start[1] << 8) | start[0])) {
    /* subsequent file */
    prgEnd += length;
    return ent;
  }
  else {
    fprintf (stderr,
	     "%u: %s: starts at 0x%04x, not 0x%04x\n",
	     lineno, name, ((uint_fast16_t) start[1] << 8) | start[0], prgEnd);
    prgStart = prgEnd = 0;
    return 0;
  }
}

/** Construct a menu name
 * @param str	name as a null-terminated character string
 * @param name	(output) the menu entry name
 */
static void
mkname (const char* str,
	octet_t* name)
{
  size_t len;
  const char* s = str;

  for (len = 0; len < sizeof (name_t); len++) {
    char c = *s++;
    if (!c)
      break;
    if (c >= 'a' && c <= 'z')
      c -= 'a' - 0x41;
    else if (c >= 'A' && c <= 'Z')
      c -= 'A' - 0xc1;
    else if (c == '\\') {
      c = *s++;
      switch (c) {
	unsigned i;
      case 'a': c = 0x07; break;
      case 'b': c = 0x08; break;
      case 't': c = 0x09; break;
      case 'n': c = 0x0a; break;
      case 'v': c = 0x0b; break;
      case 'f': c = 0x0c; break;
      case 'r': c = 0x0d; break;
      case 'x':
	/* convert up to 2 hexadecimal digits to a byte */
	c = 0;
	for (i = 0;
	     i < 2 && ((*s >= '0' && *s <= '9') ||
		       (*s >= 'a' && *s <= 'f') ||
		       (*s >= 'A' && *s <= 'F'));
	     i++) {
	  char d = *s++;
	  if (d >= '0' && d <= '9')
	    d -= '0';
	  else if (d >= 'a' && d <= 'f')
	    d -= 'a' - 0xa;
	  else if (d >= 'A' && d <= 'F')
	    d -= 'A' - 0xa;
	  c <<= 4;
	  c |= d;
	}
	if (i == 0)
	  c = 'x'; /* no digits given: translate \x to x */
	break;
      default:
	/* convert up to 3 octal digits to a byte */
	s--;
	c = 0;
	for (i = 0; *s >= '0' && *s <= '7' && i < 3; i++) {
	  c <<= 3;
	  c |= *s++ - '0';
	  if (c >= 040)
	    break; /* prevent overflow */
	}
	if (i == 0)
	  c = *s; /* no digits given: omit the \ */
	break;
      }
    }
    name[len] = c;
  }

  /* Pad with spaces */
  memset (name + len, 0x20, sizeof (name_t) - len);
}

/** Compare two menu references
 * @param a	pointer to struct menuref
 * @param b	pointer to struct menuref
 * @return	0 if a and b are equal, nonzero if not
 */
static int
refcmp (const void* a,
	const void* b)
{
  return strcmp (((const struct menuref*) a)->name,
		 ((const struct menuref*) b)->name);
}

/** Process a menu specification
 * @param file	the input stream
 */
static void
processFile (FILE* file)
{
  char* line;
  unsigned lineno;
  struct menuent menu;
  /** Menu references or labels */
  struct menuref* ref = NULL;
  /** Number of entries in ref */
  unsigned n_ref = 0;
  /** Old number of entries in ref
  (for detecting if lsearch entered the key) */
  unsigned n_ref_old;
  /** Allocated size of ref */
  unsigned ref_size = 0;

  /** menu status */
  enum {
    /** closed: menu not being constructed */
    mClosed,
    /** menu opened */
    mOpened,
    /** created: menu entry given a name but not contents */
    mCreated,
    /** populated: menu entry given a name and (one) file */
    mPopulated
  } menuStatus = mClosed;

  memset (&menu, 0, sizeof menu);

  for (lineno = 1; (line = readline (file)); lineno++) {
    struct menuref* submenu;
    struct menuref submenuref;
    char* l;

    l = strchr (line, '#');
    if (l)
      *l = 0; /* # starts a comment until the end of the line */
    while (l-- > line && isspace (*l))
      *l = 0; /* discard trailing whitespace */
    l = line;
  nextChar:
    if (*l == '(') { /* opening a menu definition */
      switch (menuStatus) {
      case mPopulated:
	if (closeEntry (&menu))
	  fprintf (stderr, "%u: no space left in menu\n", lineno);
	/* fall through */
      closeMenu:
      case mOpened:
	fprintf (stderr, "%u: closing previous menu definition\n", lineno);
	break;
      case mCreated:
	fprintf (stderr, "%u: previous menu entry is is empty\n", lineno);
	goto closeMenu;
      case mClosed:
	break;
      }

      menuStatus = mOpened;
      menuPrev = ADDR_UNKNOWN;
      l++;
      goto nextChar;
    }
    else if (*l == ')') { /* closing a menu definition */
      switch (menuStatus) {
      case mPopulated:
	if (closeEntry (&menu))
	  fprintf (stderr, "%u: no space left in menu\n", lineno);
	/* fall through */
      case mOpened:
	break;
      case mCreated:
	fprintf (stderr, "%u: last menu entry is is empty\n", lineno);
	break;
      case mClosed:
	fprintf (stderr, "%u: no menu is open\n", lineno);
	break;
      }

      menuStatus = mClosed;
      menuPrev = ADDR_UNKNOWN;
      l++;
      goto nextChar;
    }
    else if (*l == '*') { /* define a menu label: *menu_ref */
      switch (menuStatus) {
      case mPopulated:
      case mOpened:
      case mCreated:
	break;
      case mClosed:
	fprintf (stderr, "%u: no menu is open\n", lineno);
	goto nextLine;
      }

      if (n_ref >= ref_size) {
	unsigned newsize = ref_size;
	if (newsize) {
	  struct menuref* newref;
	  do newsize <<= 1; while (n_ref >= newsize);
	  newref = realloc (ref, newsize * sizeof *ref);
	  if (!newref)
	    goto oomlabel;
	  ref_size = newsize;
	  ref = newref;
	}
	else {
	  newsize = 1;
	  ref = malloc (newsize * sizeof *ref);
	  if (!ref) {
	  oomlabel:
	    fprintf (stderr, "%u: out of memory for menu label\n", lineno);
	    goto nextLine;
	  }
	  ref_size = newsize;
	}
      }

      submenuref.name = ++l;
      n_ref_old = n_ref;
      submenu = lsearch (&submenuref, ref, &n_ref, sizeof *submenu, refcmp);
      if (n_ref == n_ref_old) {
	/* Label found: Resolve deferred references to this menu. */
	struct menu* data = submenu->data;
	while (data->next) {
	  struct menu* next;

	  updateOffset (data->addr, firstAddr);

	  next = data->next;
	  free (data);
	  data = next;
	}
	/* Replace the entry. */
	data->lineno = lineno;
	data->addr = firstAddr;
	data->next = NULL;
	submenu->data = data;
      }
      else {
	/* Enter the menu. */
	submenu->name = strdup (submenu->name);
	submenu->data = malloc(sizeof *submenu->data);
	if (!submenu->name || !submenu->data) {
	  n_ref--;
	  free (submenu->name);
	  goto oomlabel;
	}
	submenu->data->next = NULL;
	submenu->data->lineno = lineno;
	submenu->data->addr = firstAddr;
      }

      goto nextLine;
    }
    else if (*l == '"') { /* "name" [3|8] filenames or "name"/menu_ref */
      char* name = l + 1;
      char* end = strchr (name, '"');
      if (end) {
	*end = 0;

	switch (menuStatus) {
	case mClosed:
	  fprintf (stderr, "%u: opening menu for \"%s\"\n",
		   lineno, name);
	  menuStatus = mOpened;
	  menuPrev = ADDR_UNKNOWN;
	  break;
	case mOpened:
	  break;
	case mCreated:
	  fprintf (stderr, "%u: predecessor of \"%s\" is empty\n",
		   lineno, name);
	  break;
	case mPopulated:
	  if (closeEntry (&menu))
	    fprintf (stderr, "%u: no space left in menu\n", lineno);
	  break;
	}

	mkname (name, menu.name);
	menuStatus = mCreated;

	l = end + 1;
	switch (*l) {
	default:
	  prgNextFlags = m0k;
	  break;
	case '3':
	  prgNextFlags = m3k, l++;
	  break;
	case '8':
	  prgNextFlags = m8k, l++;
	  break;
	case '/':
	  if (n_ref >= ref_size) {
	    unsigned newsize = ref_size;
	    if (newsize) {
	      struct menuref* newref;
	      do newsize <<= 1; while (n_ref >= newsize);
	      newref = realloc (ref, newsize * sizeof *ref);
	      if (!newref)
		goto oomref;
	      ref_size = newsize;
	      ref = newref;
	    }
	    else {
	      newsize = 1;
	      ref = malloc (newsize * sizeof *ref);
	      if (!ref) {
	      oomref:
		fprintf (stderr, "%u: out of memory for menu reference\n",
			 lineno);
		goto nextLine;
	      }
	      ref_size = newsize;
	    }
	  }

	  submenuref.name = ++l;
	  n_ref_old = n_ref;
	  submenu = lsearch (&submenuref, ref, &n_ref, sizeof *submenu, refcmp);
	  if (n_ref_old != n_ref) {
	    /* First unresolved menu reference */
	    submenu->name = strdup (submenu->name);
	    submenu->data = malloc(sizeof *submenu->data);
	    if (!submenu->name || !submenu->data) {
	    oomref2:
	      n_ref--;
	      free (submenu->name);
	      free (submenu->data);
	      goto oomref;
	    }
	    submenu->data->next = calloc (1, sizeof *submenu->data->next);
	    if (!submenu->data->next)
	      goto oomref2;

	    submenu->data->lineno = lineno;
	    submenu->data->addr = writeSubmenu (&menu, ADDR_UNKNOWN);
	    if (submenu->data->addr == ADDR_UNKNOWN) {
	      free (submenu->data->next);
	      free (submenu->name);
	      free (submenu->data);
	      n_ref--;
	      goto unresolvedMenuRef;
	    }
	  }
	  else if (submenu->data->next) {
	    /* Defer an unresolved menu reference */
	    struct menu* data = malloc (sizeof *data);
	    if (!data)
	      goto oomref;
	    data->lineno = lineno;
	    data->addr = writeSubmenu (&menu, ADDR_UNKNOWN);
	    if (data->addr == ADDR_UNKNOWN) {
	      free (data);
	    unresolvedMenuRef:
	      fprintf (stderr, "%u: unresolvable submenu reference: ", lineno);
	      goto nextLine;
	    }

	    data->next = submenu->data->next;
	    submenu->data = data;
	  }
	  else
	    /* Resolved menu reference */
	    writeSubmenu (&menu, ((struct menu*) submenu->data)->addr);
	  menuStatus = mOpened;
	  goto nextLine;
	}
	prgPrevFlags = mBASIC;

	if (isspace (*l))
	  goto addFile;
      }
      else
	fprintf (stderr, "%u: unterminated '\"'\n", lineno);
    }
    else if (isspace (*l)) {
    addFile:
      while (isspace (*l))
	l++;
      if (*l) {
	switch (menuStatus) {
	case mClosed:
	case mOpened:
	  fprintf (stderr, "%u: no menu specified\n", lineno);
	  break;
	case mPopulated:
	  if (addFile (l, lineno, &menu) && !closeEntry (&menu))
	    menuStatus = mOpened;
	  else {
	    fprintf (stderr, "%u: no space left in menu\n", lineno);
	    free (line);
	    return;
	  }
	  break;
	case mCreated:
	  menuStatus = addFile (l, lineno, &menu) ? mPopulated : mOpened;
	  break;
	}
      }
      goto nextLine;
    }
    while (isspace (*l))
      l++;
    if (*l)
      fprintf (stderr, "%u: ignoring the rest of the line: %s\n",
	       lineno, l);
  nextLine:
    free (line);
  }
  switch (menuStatus) {
  case mClosed:
    break;
  case mPopulated:
    if (closeEntry (&menu))
      fprintf (stderr, "%u: no space left in menu\n", lineno);
    /* fall through */
  case mOpened:
    fprintf (stderr, "%u: last menu left open\n", lineno);
    break;
  case mCreated:
    fprintf (stderr, "%u: last menu entry left open\n", lineno);
    break;
  }

  if (ref) {
    unsigned i;
    /* Ensure that there are no unresolved menu references */

    for (i = 0; i < n_ref; i++) {
      struct menu* data = ref[i].data;
      while (data->next) {
	struct menu* next;
	fprintf (stderr, "%u: unresolved reference to menu %s\n",
		 data->lineno, ref[i].name);

	next = data->next;
	free (data);
	data = next;
      }
      free (data);
      free (ref[i].name);
    }

    free (ref);
  }
}

/** The main program
 * @param argc	number of command-line arguments
 * @param argv	the command line arguments
 */
int
main (int argc, char** argv)
{
  if (argc != 2) {
    fputs ("usage: ", stderr);
    fputs (*argv, stderr);
    fputs (" cart.bin < cartlist.txt\n", stderr);
    return 1;
  }

  if (openCartridge (argv[1])) {
    processFile (stdin);
    fclose (cart);
    return 0;
  }

  if (cart)
    fclose (cart);
  return 2;
}
